//Microsoft Developer Studio generated resource script.
//
#define APSTUDIO_READONLY_SYMBOLS
/////////////////////////////////////////////////////////////////////////////
//
// Generated from the TEXTINCLUDE 2 resource.
//
#include <windows.h> // per https://mattshaw.org/news/fix-for-error-rc2144-primary-language-id-not-a-number/
#include "version.h"

/////////////////////////////////////////////////////////////////////////////
#undef APSTUDIO_READONLY_SYMBOLS

/////////////////////////////////////////////////////////////////////////////
// English (Canada) resources

#if !defined(AFX_RESOURCE_DLL) || defined(AFX_TARG_ENC)
#ifdef _WIN32
LANGUAGE LANG_ENGLISH, SUBLANG_ENGLISH_CAN
#pragma code_page(1252)
#endif //_WIN32

/////////////////////////////////////////////////////////////////////////////
//
// Icon
//

// Icon with lowest ID value placed first to ensure application icon
// remains consistent on all systems.
IDI_JOVE_ICON           ICON    DISCARDABLE     "jjove.ico"

#ifdef APSTUDIO_INVOKED
/////////////////////////////////////////////////////////////////////////////
//
// TEXTINCLUDE
//

1 TEXTINCLUDE DISCARDABLE 
BEGIN
    "\0"
END

2 TEXTINCLUDE DISCARDABLE 
BEGIN
    "\0"
END

3 TEXTINCLUDE DISCARDABLE 
BEGIN
    "#include ""version.h""\r\n"
    "\0"
END

#endif    // APSTUDIO_INVOKED


#ifndef _MAC
/////////////////////////////////////////////////////////////////////////////
//
// Version
//


VS_VERSION_INFO VERSIONINFO
 FILEVERSION jversion_lnum
 PRODUCTVERSION jversion_lnum
 FILEFLAGSMASK 0x3fL
#ifdef _DEBUG
 FILEFLAGS 0x1L
#else
 FILEFLAGS 0x0L
#endif
 FILEOS 0x40004L
 FILETYPE 0x1L
 FILESUBTYPE 0x0L
BEGIN
    BLOCK "StringFileInfo"
    BEGIN
        BLOCK "100904b0"
        BEGIN
            VALUE "Comments", "Originally authored by Jonathan Payne\0"
            VALUE "CompanyName", "github.com/jonmacs/jove\0"
            VALUE "LegalCopyright", "Copyright (C) 1986-1999 by Jonathan Payne."
            VALUE "Copyright",  "This program is Copyright (C) 1986-1999 by Jonathan Payne.  JOVE is "
                  "provided by Jonathan and Jovehacks without charge and without "
                  "warranty.  You may copy, modify, and/or distribute JOVE, provided that "
                  "this notice is included in all the source files and documentation.\0"
            VALUE "FileDescription", "Jonathan's Own Version of Emacs\0"
            VALUE "FileVersion", jversion "\0"
            VALUE "InternalName", "Jove\0"
            VALUE "OriginalFilename", "JOVE.EXE\0"
            VALUE "ProductName", "Jove for Win32" "\0"
            VALUE "ProductVersion", jversion "\0"
        END
    END
    BLOCK "VarFileInfo"
    BEGIN
        VALUE "Translation", 0x1009, 1200
    END
END

#endif    // !_MAC

#endif    // English (Canada) resources
/////////////////////////////////////////////////////////////////////////////



#ifndef APSTUDIO_INVOKED
/////////////////////////////////////////////////////////////////////////////
//
// Generated from the TEXTINCLUDE 3 resource.
//
#include "version.h"

/////////////////////////////////////////////////////////////////////////////
#endif    // not APSTUDIO_INVOKED
