/**************************************************************************
 * This program is Copyright (C) 1986-2002 by Jonathan Payne.  JOVE is    *
 * provided by Jonathan and Jovehacks without charge and without          *
 * warranty.  You may copy, modify, and/or distribute JOVE, provided that *
 * this notice is included in all the source files and documentation.     *
 **************************************************************************/

/* NOTE: use only spaces, no TABS below to ensure "make .version" works */
#define jversion "4.17.5.3" /* make .version uses sed to extract this */
#define jversion_lnum 4,17,5,3   /* for windows jjove.rc file */
