/**************************************************************************
 * This program is Copyright (C) 1986-2002 by Jonathan Payne.  JOVE is    *
 * provided by Jonathan and Jovehacks without charge and without          *
 * warranty.  You may copy, modify, and/or distribute JOVE, provided that *
 * this notice is included in all the source files and documentation.     *
 **************************************************************************/

/* Commands: */

extern void
	CasRegLower proto((void)),
	CasRegUpper proto((void)),
	CapChar proto((void)),
	CapWord proto((void)),
	LowWord proto((void)),
	UppWord proto((void));
